@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,6 +23,9 @@
 # Most of the guts of this file have been moved into im.py, with only some
 # of the stuff thar uses images remaining.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import renpy.display
 import renpy.text
 from renpy.display.render import render, Render
@@ -35,7 +38,21 @@ images = { }
 
 # A map from image tag to lists of possible attributes for images with that
 # tag.
-image_attributes = collections.defaultdict(list)
+image_attributes = collections.defaultdict(dict)
+
+# The set of image names Ren'Py knows about, as strings with spaces.
+image_names = [ ]
+
+
+def list_images():
+    """
+    :doc: image_func
+
+    Returns a list of images that have been added to Ren'Py, as a list of
+    strings with spaces between the name components.
+    """
+
+    return image_names
 
 
 def get_available_image_tags():
@@ -80,7 +97,7 @@ def get_tag_method(tag, method):
     it returns None.
     """
 
-    ti =  images.get((tag,), None)
+    ti = images.get((tag,), None)
 
     if ti is None:
         return None
@@ -93,16 +110,33 @@ def check_image_attributes(tag, attributes):
     :doc: image_func
 
     Checks to see if there is a unique image with the given tag and
-    attributes. If there is, returns the tag and attributes in order.
+    attributes. If there is, returns the attributes in order.
     Otherwise, returns None.
     """
 
-    ca = get_tag_method(tag, "_choose_attributes")
+    l = [ ]
 
-    if ca is not None:
-        return ca(tag, attributes, None)
+    for attrs, d in image_attributes[tag].items():
 
-    l = get_available_image_attributes(tag, attributes)
+        remainder = [ i for i in attributes if i not in attrs ]
+
+        ca = getattr(d, "_choose_attributes", None)
+
+        if ca is not None:
+
+            chosen = ca(tag, remainder, None)
+            if chosen is not None:
+                l.append(list(attrs) + list(chosen))
+
+        else:
+
+            if not remainder:
+                l.append(attrs)
+
+    # Check to see if there's an image that is exactly the one we want.
+    for i in l:
+        if len(i) == len(attributes):
+            return tuple(i)
 
     if len(l) != 1:
         return None
@@ -128,32 +162,81 @@ def get_ordered_image_attributes(tag, attributes=(), sort=None):
 
     """
 
-    la = get_tag_method(tag, "_list_attributes")
-
-    if la is not None:
-        return la(tag, attributes)
+    sequences = [ ]
 
     attrcount = collections.defaultdict(int)
     attrtotalpos = collections.defaultdict(float)
 
-    for attrlist in get_available_image_attributes(tag, attributes):
-        for i, attr in enumerate(attrlist):
-            attrcount[attr] += 1
-            attrtotalpos[attr] += i
+    for attrs, d in sorted(image_attributes[tag].items()):
+
+        la = getattr(d, "_list_attributes", None)
+        if la is not None:
+
+            sequence = list(attrs) + la(tag, [ i for i in attributes if i not in attrs ])
+
+            if not all(i in sequence for i in attributes):
+                continue
+
+            sequences.append(sequence)
+
+        else:
+
+            if not all(i in attrs for i in attributes):
+                continue
+
+            for i, attr in enumerate(attrs):
+                attrcount[attr] += 1
+                attrtotalpos[attr] += i
 
     if sort is None:
+        return list(set(attrcount.keys()) | set(j for i in sequences for j in i))
 
-        return list(attrcount.keys())
+    # If we have a sequence, do a topological sort on the before-after relation -
+    # with an ajustment to make sure it will complete even if it loops.
 
-    else:
+    rv = [ ]
 
-        l = [ ]
+    # A map from an attribute to all the attributes it is after.
+    after = collections.defaultdict(set)
 
-        for attr in attrcount:
-            l.append((attrtotalpos[attr] / attrcount[attr], sort(attr), attr))
+    for i in sequences:
+        while i:
 
-        l.sort()
-        return [ i[2] for i in l ]
+            j = i.pop(0)
+
+            # Ensure it exists.
+            after[j]
+
+            for k in i:
+                after[k].add(j)
+
+    while after:
+
+        mincount = min(len(i) for i in after.values())
+        ready = set(k for k, v in after.items() if len(v) == mincount)
+
+        for i in ready:
+            del after[i]
+
+        for k in after:
+            after[k] = after[k] - ready
+
+        ready = list(ready)
+        ready.sort(key=lambda a : (sort(a), a))
+
+        rv.extend(ready)
+
+    l = [ ]
+
+    for attr in attrcount:
+        if attr not in rv:
+            l.append((attrtotalpos[attr] // attrcount[attr], sort(attr), attr))
+
+    l.sort()
+    for i in l:
+        rv.append(i[2])
+
+    return rv
 
 
 def register_image(name, d):
@@ -169,7 +252,9 @@ def register_image(name, d):
     rest = name[1:]
 
     images[name] = d
-    image_attributes[tag].append(rest)
+    image_attributes[tag][rest] = d
+
+    image_names.append(" ".join(name))
 
 
 def image_exists(name, exact=False):
@@ -204,6 +289,20 @@ def image_exists(name, exact=False):
     return False
 
 
+def get_registered_image(name):
+    """
+    :doc: image_func
+
+    If an image with the same name has been registered with renpy.register_image,
+    returns it. Otherwise, returns None.
+    """
+
+    if not isinstance(name, tuple):
+        name = tuple(name.split())
+
+    return images.get(name)
+
+
 def wrap_render(child, w, h, st, at):
     rend = render(child, w, h, st, at)
     rv = Render(rend.width, rend.height)
@@ -298,10 +397,15 @@ class ImageReference(renpy.display.core.Displayable):
             error("Image '%s' not found." % ' '.join(self.name))
             return False
 
+        if name and (self._args.name == name):
+            error("Image '{}' refers to itself.".format(' '.join(name)))
+            return False
+
+        args += self._args.args
+
         try:
 
             a = self._args.copy(name=name, args=args)
-
             self.target = target._duplicate(a)
 
         except Exception as e:
@@ -310,6 +414,7 @@ class ImageReference(renpy.display.core.Displayable):
                 raise
 
             error(str(e))
+            return False
 
         # Copy the old transform over.
         new_transform = self.target._target()
@@ -329,6 +434,9 @@ class ImageReference(renpy.display.core.Displayable):
 
     def _duplicate(self, args):
 
+        if args and args.args:
+            args.extraneous()
+
         rv = self._copy(args)
         rv.target = None
 
@@ -362,6 +470,12 @@ class ImageReference(renpy.display.core.Displayable):
         rv.target = target
         return rv
 
+    def _handles_event(self, event):
+        if self.target is None:
+            return False
+
+        return self.target._handles_event(event)
+
     def _hide(self, st, at, kind):
         if self.target is None:
             self.find_target()
@@ -437,17 +551,14 @@ class DynamicImage(renpy.display.core.Displayable):
     # Have we been locked, so we never change?
     locked = False
 
+    # The name used for hashing.
+    hash_name = None
+
     def __init__(self, name, scope=None, **properties):
         super(DynamicImage, self).__init__(**properties)
 
         self.name = name
 
-        if scope is not None:
-            self.find_target(scope)
-            self._uses_scope = True
-        else:
-            self._uses_scope = False
-
         if isinstance(name, basestring) and ("[prefix_" in name):
             self._duplicatable = True
 
@@ -456,6 +567,12 @@ class DynamicImage(renpy.display.core.Displayable):
                 if ("[prefix_" in i):
                     self._duplicatable = True
 
+        if scope is not None:
+            self.find_target(scope)
+            self._uses_scope = True
+        else:
+            self._uses_scope = False
+
     def _scope(self, scope, update):
         return self.find_target(scope, update)
 
@@ -463,7 +580,14 @@ class DynamicImage(renpy.display.core.Displayable):
         return u"DynamicImage {!r}".format(self.name)
 
     def __hash__(self):
-        return hash(self.name)
+
+        if self.hash_name is None:
+            if isinstance(self.name, list):
+                self.hash_name = tuple(self.name)
+            else:
+                self.hash_name = self.name
+
+        return hash(self.hash_name)
 
     def __eq__(self, o):
         if self is o:
@@ -475,6 +599,9 @@ class DynamicImage(renpy.display.core.Displayable):
         if self.name != o.name:
             return False
 
+        if self._uses_scope and (self.target != o.target):
+            return False
+
         return True
 
     def _target(self):
@@ -483,26 +610,44 @@ class DynamicImage(renpy.display.core.Displayable):
         else:
             return self
 
+    def set_style_prefix(self, prefix, root):
+
+        if (prefix != self.style.prefix) and self._duplicatable:
+            self.target = None
+            self.raw_target = None
+
+        super(DynamicImage, self).set_style_prefix(prefix, root)
+
     def find_target(self, scope=None, update=True):
 
         if self.locked and (self.target is not None):
             return
 
         if self._args.prefix is None:
-            prefix = ""
+            if self._duplicatable:
+                prefix = self.style.prefix
+            else:
+                prefix = ""
         else:
             prefix = self._args.prefix
 
         try:
-            target = renpy.easy.dynamic_image(self.name, scope, prefix=prefix)
+            search = [ ]
+            target = renpy.easy.dynamic_image(self.name, scope, prefix=prefix, search=search)
+        except KeyError as ke:
+            raise Exception("In DynamicImage %r: Could not find substitution '%s'." % (self.name, str(ke.args[0])))
         except Exception as e:
             raise Exception("In DynamicImage %r: %r" % (self.name, e))
 
         if target is None:
-            error = "DynamicImage %r: did not resolve to an image." % (self.name,)
+            error = "DynamicImage %r: could not find image." % (self.name,)
 
-            if self._args.prefix:
-                error += " prefix=" + self._args.prefix
+            if len(search) == 1:
+                error += " (%r)" % (search[0],)
+            elif len(search) == 2:
+                error += " (%r, %r)" % (search[0], search[1])
+            elif len(search) > 2:
+                error += " (%r, %r, and %d more.)" % (search[0], search[1], len(search) - 2)
 
             raise Exception(error)
 
@@ -536,8 +681,13 @@ class DynamicImage(renpy.display.core.Displayable):
         return True
 
     def _duplicate(self, args):
+
+        if args and args.args:
+            args.extraneous()
+
         rv = self._copy(args)
         rv.target = None
+        rv.raw_target = None
         # This does not set _duplicatable, since it should always remain the
         # same.
         return rv
@@ -551,6 +701,12 @@ class DynamicImage(renpy.display.core.Displayable):
         rv.locked = True
         return rv
 
+    def _handles_event(self, event):
+        if self.target is None:
+            return False
+
+        return self.target._handles_event(event)
+
     def _hide(self, st, at, kind):
         if self.target is None:
             self.find_target()
@@ -640,13 +796,16 @@ class ShownImageInfo(renpy.object.Object):
                     self.attributes[layer, tag] = self.images[layer][tag][1:]
                     self.shown.add((layer, tag))
 
-    def get_attributes(self, layer, tag):
+    def get_attributes(self, layer, tag, default=()):
         """
         Get the attributes associated the image with tag on the given
         layer.
         """
 
-        return self.attributes.get((layer, tag), ())
+        if layer is None:
+            layer = renpy.config.tag_layer.get(tag, "master")
+
+        return self.attributes.get((layer, tag), default)
 
     def showing(self, layer, name, exact=False):
         """
@@ -657,6 +816,9 @@ class ShownImageInfo(renpy.object.Object):
         tag = name[0]
         rest = name[1:]
 
+        if layer is None:
+            layer = renpy.config.tag_layer.get(tag, "master")
+
         if (layer, tag) not in self.shown:
             return None
 
@@ -681,6 +843,14 @@ class ShownImageInfo(renpy.object.Object):
 
         return { t for l, t in self.shown if l == layer }
 
+    def get_hidden_tags(self, layer):
+        """
+        Returns the set of tags on layer that have attributes,
+        but aren't being shown.
+        """
+
+        return { t for l, t in self.attributes if l == layer if (l, t) not in self.shown }
+
     def predict_scene(self, layer):
         """
         Predicts the scene statement being called on layer.
@@ -689,7 +859,7 @@ class ShownImageInfo(renpy.object.Object):
         if layer is None:
             layer = 'master'
 
-        for l, t in self.attributes.keys():
+        for l, t in list(self.attributes.keys()):
             if l == layer:
                 del self.attributes[l, t]
 
@@ -707,6 +877,9 @@ class ShownImageInfo(renpy.object.Object):
         tag = name[0]
         rest = name[1:]
 
+        if layer is None:
+            layer = renpy.config.tag_layer.get(tag, "master")
+
         self.attributes[layer, tag] = rest
 
         if show:
@@ -715,6 +888,9 @@ class ShownImageInfo(renpy.object.Object):
     def predict_hide(self, layer, name):
         tag = name[0]
 
+        if layer is None:
+            layer = renpy.config.tag_layer.get(tag, "master")
+
         if (layer, tag) in self.attributes:
             del self.attributes[layer, tag]
 
@@ -724,49 +900,76 @@ class ShownImageInfo(renpy.object.Object):
         """
         Given a layer, tag, and an image name (with attributes),
         returns the canonical name of an image, if one exists. Raises
-        an exception if it's ambiguious, and returns None if an image
+        an exception if it's ambiguous, and returns None if an image
         with that name couldn't be found.
         """
 
+        f = renpy.config.adjust_attributes.get(name[0], None) or renpy.config.adjust_attributes.get(None, None)
+        if f is not None:
+            name = f(name)
+
+        if layer is None:
+            layer = renpy.config.tag_layer.get(tag, "master")
+
         # If the name matches one that exactly exists, return it.
         if (name in images) and not (wanted or remove):
-            return name
+            ca = getattr(images[name], "_choose_attributes", None)
+
+            if ca is None:
+                return name
 
         nametag = name[0]
 
-        # The set of attributes a matching image must have.
-        required = set(name[1:])
-
         # The set of attributes a matching image may have.
-        optional = set(wanted) | set(self.attributes.get((layer, tag), [ ]))
+        optional = list(wanted) + list(self.attributes.get((layer, tag), [ ]))
+
+        # The list of attributes a matching image must have.
+        required = [ ]
 
-        # Deal with banned attributes..
         for i in name[1:]:
             if i[0] == "-":
-                optional.discard(i[1:])
-                required.discard(i)
+
+                i = i[1:]
+
+                if i in optional:
+                    optional.remove(i)
+
+                if i in required:
+                    required.remove(i)
+
+            else:
+                required.append(i)
 
         for i in remove:
-            optional.discard(i)
+            if i in optional:
+                optional.remove(i)
 
         return self.choose_image(nametag, required, optional, name)
 
     def choose_image(self, tag, required, optional, exception_name):
 
-        ca = get_tag_method(tag, "_choose_attributes")
-        if ca is not None:
-            attrs = ca(tag, required, optional)
-
-            if attrs is not None:
-                return (tag,) + attrs
-
         # The longest length of an image that matches.
         max_len = -1
 
         # The list of matching images.
         matches = None
 
-        for attrs in image_attributes[tag]:
+        for attrs, d in image_attributes[tag].items():
+
+            if not all((i in required) or (i in optional) for i in attrs):
+                continue
+
+            ca = getattr(d, "_choose_attributes", None)
+
+            if ca:
+                ca_required = [ i for i in required if i not in attrs ]
+                ca_optional = [ i for i in optional if i not in attrs ]
+                newattrs = ca(tag, ca_required, ca_optional)
+
+                if newattrs is None:
+                    continue
+
+                attrs = attrs + newattrs
 
             num_required = 0
 
@@ -775,9 +978,6 @@ class ShownImageInfo(renpy.object.Object):
                     num_required += 1
                     continue
 
-                elif i not in optional:
-                    break
-
             else:
 
                 # We don't have any not-found attributes. But we might not
@@ -795,7 +995,7 @@ class ShownImageInfo(renpy.object.Object):
                     max_len = len_attrs
                     matches = [ ]
 
-                matches.append((tag, ) + attrs)
+                matches.append((tag,) + attrs)
 
         if matches is None:
             return None
@@ -808,8 +1008,8 @@ class ShownImageInfo(renpy.object.Object):
         else:
             return None
 
-renpy.display.core.ImagePredictInfo = ShownImageInfo
 
+renpy.display.core.ImagePredictInfo = ShownImageInfo
 
 # Functions that have moved from this module to other modules,
 # that live here for the purpose of backward-compatibility.
